package de.fhdw.wtf.common.ast;

import de.fhdw.wtf.common.ast.visitor.NameReturnVisitor;
import de.fhdw.wtf.common.ast.visitor.NameVisitor;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnBoolean;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnName;
import de.fhdw.wtf.common.ast.visitor.NameVisitorReturnNameException;
import de.fhdw.wtf.common.exception.referencer.InvalidTypeReferenceException;
import de.fhdw.wtf.common.token.IdentifierToken;
import de.fhdw.wtf.common.token.Token;

/**
 * This class represents a {@link UnqualifiedName}. A {@link Name} is unqualified if contains just the name of a
 * {@link GroupElement}.
 * 
 */
public final class UnqualifiedName implements Name {
	
	/**
	 * generated.
	 */
	private static final long serialVersionUID = 1489081265226782458L;
	
	/**
	 * Identifier-Token.
	 */
	private final IdentifierToken identifierToken;
	
	/**
	 * This for Multiple Inheritance.
	 */
	private final SyntaxObjectInterface thiis;
	
	/**
	 * Constructor of {@link UnqualifiedName}.
	 * 
	 * @param identifierToken
	 *            identifierToken
	 * @param lastToken
	 *            lastToken
	 */
	private UnqualifiedName(final IdentifierToken identifierToken, final Token lastToken) {
		this.identifierToken = identifierToken;
		this.thiis = SyntaxObjectConcret.create(identifierToken, lastToken);
	}
	
	/**
	 * Creates a {@link UnqualifiedName}-Object.
	 * 
	 * @param identifierToken
	 *            identifierToken
	 * @param lastToken
	 *            lastToken
	 * @return The UnqualifiedName-Object
	 */
	public static UnqualifiedName create(final IdentifierToken identifierToken, final Token lastToken) {
		return new UnqualifiedName(identifierToken, lastToken);
	}
	
	/**
	 * Creates a {@link UnqualifiedName}-Object.
	 * 
	 * @param identifierToken
	 *            identifierToken
	 * @return The UnqualifiedName-Object
	 */
	public static UnqualifiedName create(final IdentifierToken identifierToken) {
		return new UnqualifiedName(identifierToken, null);
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof UnqualifiedName) {
			final UnqualifiedName other = (UnqualifiedName) o;
			return this.identifierToken.stringRepresentation()
					.equals(other.getIdentifierToken().stringRepresentation());
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return this.identifierToken.stringRepresentation().hashCode();
	}
	
	@Override
	public Name addName(final IdentifierToken identifier) {
		return this.addName(identifier, null);
	}
	
	@Override
	public Name addName(final IdentifierToken identifier, final Token lastToken) {
		final UnqualifiedName rest = UnqualifiedName.create(identifier, lastToken);
		return QualifiedName.create(this, rest, identifier, lastToken);
	}
	
	@Override
	public String toString() {
		return this.getIdentifierToken().stringRepresentation();
	}
	
	@Override
	public Name visit(final NameVisitorReturnName v) {
		return v.handleUnqualifiedName(this);
	}
	
	@Override
	public Name concat(final Name name) {
		return name.visit(new NameVisitorReturnName() {
			
			@Override
			public Name handleUnqualifiedName(final UnqualifiedName unqName) {
				return UnqualifiedName.this.addName(unqName.getIdentifierToken());
			}
			
			@Override
			public Name handleQualifiedName(final QualifiedName qName) {
				Name result = UnqualifiedName.this.addName(qName.getFirst().getIdentifierToken());
				result = result.concat(qName.getRest());
				return result;
			}
		});
	}
	
	@Override
	public <X> X visit(final NameReturnVisitor<X> v) {
		return v.handle(this);
	}
	
	@Override
	public UnqualifiedName getLastAddedName() {
		return this;
	}
	
	@Override
	public boolean equalsPartial(final Name name) {
		return name.visit(new NameVisitorReturnBoolean() {
			
			@Override
			public boolean handleUnqualifiedName(final UnqualifiedName unqName) {
				return UnqualifiedName.this.equals(unqName);
			}
			
			@Override
			public boolean handleQualifiedName(final QualifiedName qName) {
				return UnqualifiedName.this.equals(qName.getFirst());
			}
		});
	}
	
	@Override
	public void visit(final NameVisitor v) throws InvalidTypeReferenceException {
		v.handle(this);
	}
	
	@Override
	public boolean visit(final NameVisitorReturnBoolean v) {
		return v.handleUnqualifiedName(this);
	}
	
	@Override
	public Name visit(final NameVisitorReturnNameException<InvalidTypeReferenceException> v)
			throws InvalidTypeReferenceException {
		return v.handleUnqualifiedName(this);
	}
	
	// // PROJECTIONS ////
	
	/**
	 * Get the Identifier-Token.
	 * 
	 * @return Identifier-Token
	 */
	public IdentifierToken getIdentifierToken() {
		return this.identifierToken;
	}
	
	@Override
	public Token getFirstToken() {
		return this.getThis().getFirstToken();
	}
	
	@Override
	public void setFirstToken(final Token firstToken) {
		this.getThis().setFirstToken(firstToken);
	}
	
	@Override
	public Token getLastToken() {
		return this.getThis().getLastToken();
	}
	
	@Override
	public void setLastToken(final Token lastToken) {
		this.getThis().setLastToken(lastToken);
	}
	
	/**
	 * Returned the This of the Multiple Inheritance.
	 * 
	 * @return this
	 */
	private SyntaxObjectInterface getThis() {
		return this.thiis;
	}
	
}
