package de.fhdw.wtf.common.ast.type;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import de.fhdw.wtf.common.ast.visitor.AtomicTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitor;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorException;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.CompositeTypeVisitorReturnException;
import de.fhdw.wtf.common.ast.visitor.TypeVisitor;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturn;
import de.fhdw.wtf.common.ast.visitor.TypeVisitorReturnException;
import de.fhdw.wtf.common.exception.ast.ASTException;
import de.fhdw.wtf.common.exception.parser.SumIsAnythingException;
import de.fhdw.wtf.common.token.DummyToken;
import de.fhdw.wtf.common.token.Token;
import de.fhdw.wtf.common.token.symbols.SymbolConstants;

/**
 * A {@link ThrownType} represents a Type that should be thrown. It contains an underlying Type, which is the reference
 * to the type that should be thrown.
 */
public class ThrownType extends CompositeType {
	
	/**
	 * The string that represents the error message "A list can not be thrown.".
	 */
	private static final String LIST_NOT_THROWABLE_EXCEPTION_MESSAGE = "A list can not be thrown.";
	
	/**
	 * The string that represents the error message "A map can not be thrown.".
	 */
	private static final String MAP_NOT_THROWABLE_EXCEPTION_MESSAGE = "A map can not be thrown.";
	
	/**
	 * The string that represents the error message "A product can not be thrown.".
	 */
	private static final String PRODUCT_NOT_THROWABLE_EXCEPTION_MESSAGE = "A product can not be thrown.";
	
	/**
	 * The type that should be thrown.
	 */
	private final Type underlyingType;
	
	/**
	 * The constructor of {@link ThrownType}. It defines the firstToken and lastToken and the underlying Type, that
	 * should be thrown.
	 * 
	 * @param firstToken
	 *            the token of the token stream, where this thrown type begins.
	 * @param underlyingType
	 *            the throwing type.
	 * @param lastToken
	 *            the token of the token stream, where this thrown type ends.
	 */
	protected ThrownType(final Token firstToken, final Type underlyingType, final Token lastToken) {
		super(firstToken, lastToken);
		this.underlyingType = underlyingType;
	}
	
	/**
	 * Creates a new Object of {@link ThrownType} with the firstToken, lastToken and the underlyingType, that should be
	 * thrown.
	 * 
	 * @param firstToken
	 *            the token of the token stream, where this thrown type begins.
	 * @param underlyingType
	 *            the throwing type.
	 * @param lastToken
	 *            the token of the token stream, where this thrown type ends.
	 * @return ThrownType
	 */
	public static ThrownType create(final Token firstToken, final Type underlyingType, final Token lastToken) {
		return new ThrownType(firstToken, underlyingType, lastToken);
	}
	
	/**
	 * Creates a new Object of {@link ThrownType} with the firstToken and the underlyingType, that should be thrown. The
	 * lastToken will be set to null.
	 * 
	 * @param firstToken
	 *            the token of the token stream, where this thrown type begins.
	 * @param underlyingType
	 *            the throwing type.
	 * @return ThrownType
	 */
	public static ThrownType create(final Token firstToken, final Type underlyingType) {
		return new ThrownType(firstToken, underlyingType, null);
	}
	
	/**
	 * Return the underlying Type, that should be thrown.
	 * 
	 * @return Type
	 */
	public Type getUnderlyingType() {
		return this.underlyingType;
	}
	
	@Override
	public int compareTo(final Type o) {
		return o.accept(new TypeVisitorReturn<Integer>() {
			
			@Override
			public Integer handle(final AtomicType atomicType) {
				return atomicType.accept(new AtomicTypeVisitorReturn<Integer>() {
					
					@Override
					public Integer handle(final BaseType baseType) {
						return 1;
					}
					
					@Override
					public Integer handle(final ClassType clazz) {
						return -1;
					}
				});
			}
			
			@Override
			public Integer handle(final CompositeType compositeType) {
				return compositeType.accept(new CompositeTypeVisitorReturn<Integer>() {
					
					@Override
					public Integer handle(final ListType list) {
						return -1;
					}
					
					@Override
					public Integer handle(final MapType map) {
						return -1;
					}
					
					@Override
					public Integer handle(final ProductType product) {
						return -1;
					}
					
					@Override
					public Integer handle(final SumType sum) {
						return -1;
					}
					
					@Override
					public Integer handle(final ThrownType thrownType) {
						return ThrownType.this.getUnderlyingType().compareTo(thrownType.getUnderlyingType());
					}
				});
			}
			
			@Override
			public Integer handle(final TypeProxy typeProxy) {
				return 1;
			}
		});
	}
	
	@Override
	public void accept(final CompositeTypeVisitor v) {
		v.handle(this);
	}
	
	@Override
	public <X> X accept(final CompositeTypeVisitorReturn<X> v) {
		return v.handle(this);
	}
	
	@Override
	public <X extends Exception> void accept(final CompositeTypeVisitorException<X> v) throws X {
		v.handle(this);
	}
	
	@Override
	public <X, Y extends Exception> X accept(final CompositeTypeVisitorReturnException<X, Y> v) throws Y {
		return v.handle(this);
	}
	
	@Override
	public String getTypeString() {
		return this.getUnderlyingType().getTypeString() + SymbolConstants.EXCLAMATION;
	}
	
	@Override
	public String toString() {
		return this.getUnderlyingType().toString() + SymbolConstants.EXCLAMATION;
	}
	
	@Override
	public boolean equals(final Object obj) {
		if (obj instanceof ThrownType) {
			final ThrownType thrownType = (ThrownType) obj;
			return this.getUnderlyingType().equals(thrownType.getUnderlyingType());
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return this.getUnderlyingType().hashCode() + SymbolConstants.EXCLAMATION.hashCode();
	}
	
	@Override
	public boolean isSupertypeOf(final Type type) {
		return false;
	}
	
	@Override
	public Collection<Type> getSuperTypes() {
		return new ArrayList<>();
	}
	
	@Override
	public Type standardize() throws SumIsAnythingException {
		final Type standardizedType = this.getUnderlyingType().standardize();
		
		return standardizedType.accept(new TypeVisitorReturnException<Type, SumIsAnythingException>() {
			
			@Override
			public Type handle(final TypeProxy typeProxy) throws SumIsAnythingException {
				return ThrownType.create(DummyToken.getInstance(), typeProxy, DummyToken.getInstance());
			}
			
			@Override
			public Type handle(final CompositeType compositeType) throws SumIsAnythingException {
				return compositeType.accept(new CompositeTypeVisitorReturnException<Type, SumIsAnythingException>() {
					
					@Override
					public Type handle(final ListType list) throws SumIsAnythingException {
						throw new ASTException(ThrownType.LIST_NOT_THROWABLE_EXCEPTION_MESSAGE);
					}
					
					@Override
					public Type handle(final MapType map) throws SumIsAnythingException {
						throw new ASTException(ThrownType.MAP_NOT_THROWABLE_EXCEPTION_MESSAGE);
					}
					
					@Override
					public Type handle(final ProductType product) throws SumIsAnythingException {
						throw new ASTException(ThrownType.PRODUCT_NOT_THROWABLE_EXCEPTION_MESSAGE);
					}
					
					@Override
					public Type handle(final SumType sum) throws SumIsAnythingException {
						
						final List<Type> copy = new ArrayList<>();
						
						copy.addAll(sum.getElements());
						final List<Type> newElements = new ArrayList<>();
						sum.getElements().clear();
						
						final Iterator<Type> iterator = copy.iterator();
						while (iterator.hasNext()) {
							final Type currentSubType = iterator.next();
							
							currentSubType.accept(new TypeVisitor() {
								
								@Override
								public void handle(final TypeProxy typeProxy) {
									newElements.add(ThrownType.create(
											DummyToken.getInstance(),
											typeProxy,
											DummyToken.getInstance()));
								}
								
								@Override
								public void handle(final CompositeType compositeType2) {
									compositeType2.accept(new CompositeTypeVisitor() {
										
										@Override
										public void handle(final ListType list) {
											throw new ASTException(ThrownType.LIST_NOT_THROWABLE_EXCEPTION_MESSAGE);
										}
										
										@Override
										public void handle(final MapType map) {
											throw new ASTException(ThrownType.MAP_NOT_THROWABLE_EXCEPTION_MESSAGE);
										}
										
										@Override
										public void handle(final ProductType product) {
											throw new ASTException(ThrownType.PRODUCT_NOT_THROWABLE_EXCEPTION_MESSAGE);
										}
										
										@Override
										public void handle(final SumType sum2) {
											throw new ASTException(
													"SumTypes can not be in sumTypes after normalisation.");
										}
										
										@Override
										public void handle(final ThrownType thrownType) {
											newElements.add(thrownType);
										}
									});
								}
								
								@Override
								public void handle(final AtomicType atomicType) {
									newElements.add(ThrownType.create(
											DummyToken.getInstance(),
											atomicType,
											DummyToken.getInstance()));
								}
							});
						}
						
						if (newElements.size() == 1) {
							return newElements.get(0);
						}
						
						Collections.sort(newElements);
						
						final Iterator<Type> sumElementsIterator = newElements.iterator();
						while (sumElementsIterator.hasNext()) {
							final Type current = sumElementsIterator.next();
							
							if (!sum.getElements().contains(current)) {
								sum.getElements().add(current);
							}
						}
						
						return sum;
					}
					
					@Override
					public Type handle(final ThrownType thrownType) throws SumIsAnythingException {
						return thrownType;
					}
				});
			}
			
			@Override
			public Type handle(final AtomicType atomicType) throws SumIsAnythingException {
				return ThrownType.create(DummyToken.getInstance(), atomicType, DummyToken.getInstance());
			}
		});
	}
	
	@Override
	public boolean isThrownType() {
		return true;
	}
}
