package de.fhdw.wtf.common.token;

import java.io.Serializable;

/**
 * The position is the begin position of a token in the original input document.
 */
public final class Position implements Serializable {
	
	/**
	 * Serial Version UID.
	 */
	static final long serialVersionUID = 238623487L;
	
	/**
	 * Line-Number in the model.
	 */
	private final int lineNumber;
	
	/**
	 * Position in the model.
	 */
	private final int position;
	
	/**
	 * Column-Number in the model.
	 * 
	 */
	private final int columnNumber;
	
	/**
	 * FilePath.
	 */
	private final String filePath;
	
	/**
	 * Constructor for {@link Position}.
	 * 
	 * @param filePath
	 *            initial file path
	 * @param lineNumber
	 *            initial line number
	 * @param position
	 *            start position of this comment in the original input
	 * @param columnNumber
	 *            initial column number
	 */
	private Position(final String filePath, final int lineNumber, final int columnNumber, final int position) {
		this.lineNumber = lineNumber;
		this.filePath = filePath;
		this.columnNumber = columnNumber;
		this.position = position;
	}
	
	/**
	 * Factory for a {@link Position}.
	 * 
	 * @param filePath
	 *            initial file path
	 * @param lineNumber
	 *            initial line number
	 * @param position
	 *            start position in the original input
	 * @param columnNumber
	 *            initial column number
	 * @return a new instance of this position
	 */
	public static Position create(final String filePath,
			final int lineNumber,
			final int columnNumber,
			final int position) {
		return new Position(filePath, lineNumber, columnNumber, position);
	}
	
	/**
	 * @return Returns the line number of a token in the original input document. Counting starts at 1.
	 */
	public int getLineNumber() {
		return this.lineNumber;
	}
	
	/**
	 * @return Returns the position of a token in the document. (Zero-based counting, Control sequences are counted as
	 *         one character)
	 */
	public int getPosition() {
		return this.position;
	}
	
	/**
	 * @return Returns the position of a token in the line. The line is given in {@link #getLineNumber()
	 *         getLineNumber()} Counting starts at 1.
	 */
	public int getColumnNumber() {
		return this.columnNumber;
	}
	
	/**
	 * @return Returns the file path of a token.
	 */
	public String getFilePath() {
		return this.filePath;
	}
	
	@Override
	public String toString() {
		return "[FilePath:" + this.getFilePath() + ",LineNumber:" + this.getLineNumber() + ",ColumnNumber:"
				+ this.getColumnNumber() + ",Position:" + this.getPosition() + "]";
	}
	
	@Override
	public boolean equals(final Object o) {
		if (o instanceof Position) {
			final Position p = (Position) o;
			return this.getFilePath().equals(p.getFilePath()) && this.getLineNumber() == p.getLineNumber()
					&& this.getPosition() == p.getPosition() && this.getColumnNumber() == p.getColumnNumber();
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return this.getFilePath().hashCode() ^ this.getLineNumber() ^ this.getPosition() ^ this.getColumnNumber();
	}
	
	/**
	 * @return creates a non-existent position object
	 */
	public static Position createDummyPosition() {
		return Position.create("", -1, -1, -1);
	}
}
