package de.fhdw.wtf.parser;

import java.util.Collection;

import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.Name;
import de.fhdw.wtf.common.ast.UnqualifiedName;
import de.fhdw.wtf.common.exception.parser.AbstractParserException;
import de.fhdw.wtf.common.stream.TokenStream;
import de.fhdw.wtf.common.token.IdentifierToken;
import de.fhdw.wtf.common.token.Token;

/**
 * Parser to parse a model from a given TokenStream.
 */
final class ModelParser {
	/**
	 * TokenStream.
	 */
	private final TokenStream stream;
	
	/**
	 * Collection of exceptions.
	 */
	private final Collection<AbstractParserException> exceptions;
	
	/**
	 * Constructor of {@link ModelParser}.
	 * 
	 * @param stream
	 *            tokenStream
	 * @param exceptions
	 *            collection of exceptions
	 */
	private ModelParser(final TokenStream stream, final Collection<AbstractParserException> exceptions) {
		this.stream = stream;
		this.exceptions = exceptions;
	}
	
	/**
	 * FactoryMethod for {@link ModelParser}.
	 * 
	 * @param stream
	 *            tokenStream
	 * @param exceptions
	 *            collection of exceptions
	 * @return The {@link ModelParser}-Object.
	 */
	static ModelParser create(final TokenStream stream, final Collection<AbstractParserException> exceptions) {
		return new ModelParser(stream, exceptions);
	}
	
	/**
	 * Parses Model from a given TokenStream. If the Stream does not contain any Token a empty Model will be returned.
	 * Any Exceptions ocurring are being put in a Collection, accessible through the Parser-Object.
	 * 
	 * @return Model
	 */
	Model parse() {
		final Model model = Model.create(this.stream.peek());
		
		while (this.stream.hasNext() && !this.stream.peek().isEndToken()) {
			try {
				final IdentifierToken identifierToken = ParserUtils.requireAndRemoveIdentifier(this.stream);
				final Token nextToken = this.stream.peek();
				ParserUtils.requireAndRemoveColonToken(this.stream);
				ParserUtils.requireAndRemoveGroupToken(this.stream);
				final Name name = UnqualifiedName.create(identifierToken, nextToken);
				model.addGroup(GroupParser.create(this.stream, this.exceptions).parse(name, identifierToken));
			} catch (final AbstractParserException e) {
				this.exceptions.add(e);
				ParserUtils.skipToSemicolonToken(this.stream);
				continue;
			}
		}
		
		model.setLastToken(this.stream.peek());
		
		return model;
	}
	
}
