package de.fhdw.wtf.walker.walker;

import de.fhdw.wtf.common.ast.Attribute;
import de.fhdw.wtf.common.ast.Constructor;
import de.fhdw.wtf.common.ast.Group;
import de.fhdw.wtf.common.ast.GroupElement;
import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.ast.Operation;
import de.fhdw.wtf.common.ast.type.ClassType;
import de.fhdw.wtf.common.ast.visitor.GroupElementExceptionVisitor;
import de.fhdw.wtf.common.exception.walker.TaskException;
import de.fhdw.wtf.common.task.DependencyTask;
import de.fhdw.wtf.common.task.TaskExecutor;
import de.fhdw.wtf.common.task.result.OKTaskResult;

/**
 * A task does its specific work, while traversing through the abstract syntax tree.
 * 
 */
public abstract class SimpleWalkerTask extends WalkerTask {
	
	/**
	 * Abstract syntax tree model, that is traversed in this task.
	 */
	private final Model m;
	
	/**
	 * Constructor for {@link SimpleWalkerTask}.
	 * 
	 * @param m
	 *            Model
	 * @param taskmanager
	 *            Taskmanager
	 */
	public SimpleWalkerTask(final Model m, final TaskExecutor taskmanager) {
		super(taskmanager);
		this.m = m;
	}
	
	@Override
	public de.fhdw.wtf.common.task.result.TaskResult doWork() {
		try {
			this.beginTask();
			this.walk();
			this.finalizeTask();
		} catch (final TaskException e) {
			return new de.fhdw.wtf.common.task.result.ExceptionalTaskResult(e);
		}
		return new OKTaskResult();
	}
	
	/**
	 * "Walks" down the tree, calling the current service at all elements.
	 * 
	 * @throws TaskException
	 *             when the implementing task throws unfixable errors
	 */
	private void walk() throws TaskException {
		for (final Group g : this.m.getGroups()) {
			this.handleGroup(g);
			this.processGroupElements(g);
		}
	}
	
	/**
	 * Walks over and processes GroupElements. Calls the abstract handle-operations of {@link WalkerTask} when needed.
	 * 
	 * @param g
	 *            : common.ast.Group.
	 * @throws TaskException
	 *             : When the implementing task throws unfixable errors
	 */
	void processGroupElements(final Group g) throws TaskException {
		for (final GroupElement element : g.getGroupElements()) {
			element.accept(new GroupElementExceptionVisitor<TaskException>() {
				@Override
				public void handle(final Group g2) throws TaskException {
					SimpleWalkerTask.this.handleGroup(g2);
					SimpleWalkerTask.this.processGroupElements(g2);
				}
				
				@Override
				public void handle(final ClassType clss) throws TaskException {
					SimpleWalkerTask.this.handleClass(clss);
					for (final Attribute a : clss.getAttributes()) {
						SimpleWalkerTask.this.handleAttribute(a, clss);
					}
					for (final Operation o : clss.getOperations()) {
						SimpleWalkerTask.this.handleConstructorOrOperation(o, clss);
					}
					for (final Constructor constructor : clss.getConstructors()) {
						SimpleWalkerTask.this.handleConstructorOrOperation(constructor, clss);
					}
				}
			});
		}
	}
	
	@Override
	public boolean containsTransitive(final DependencyTask a) {
		return false;
	}
	
	/**
	 * Returns the AST-Model of this SimpleWalkerTask.
	 * 
	 * @return Model
	 */
	public Model getModel() {
		return this.m;
	}
	
}
